__version__ = "0.0.0"

import os
import sys
from datetime import datetime

from .logger import WranglerLogger, setupLogging
from .projectcard import ProjectCard
from .roadwaynetwork import RoadwayNetwork
from .transitnetwork import TransitNetwork
from .scenario import Scenario
from .scenario import net_to_mapbox
from .utils import point_df_to_geojson, link_df_to_json, make_slug
from .utils import parse_time_spans, offset_location_reference, haversine_distance
from .utils import create_unique_shape_id
from .utils import create_location_reference_from_nodes
from .utils import create_line_string


__all__ = [
    "WranglerLogger",
    "setupLogging",
    "ProjectCard",
    "RoadwayNetwork",
    "point_df_to_geojson",
    "TransitNetwork",
    "Scenario",
    "link_df_to_json",
    "make_slug",
    "parse_time_spans",
    "offset_location_reference",
    "haversine_distance",
    "create_unique_shape_id",
    "create_location_reference_from_nodes",
    "create_line_string",
    "net_to_mapbox",
]

setupLogging(
    log_filename=os.path.join(
        os.getcwd(),
        "network_wrangler_{}.log".format(datetime.now().strftime("%Y_%m_%d__%H_%M_%S")),
    ),
    log_to_console=True,
)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

# Get the project root dir, which is the parent dir of this
cwd = os.getcwd()
project_root = os.path.dirname(cwd)

# Insert the project root dir as the first element in the PYTHONPATH.
# This lets us ensure that the source package is imported, and that its
# version is used.
sys.path.insert(0, project_root)

import network_wrangler


# -- Project information -----------------------------------------------------

project = "Network Wrangler"
copyright = "2020, Metropolitan Council"
author = "Metropolitan Council"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "recommonmark",
    "sphinx.ext.todo",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# -- Options for Autodoc -------------------------------------------------
autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "inherited-members": True,
    "imported-members": True,
    "show-inheritance": True,
    "member-order": "groupwise",
}

autoclass_content = "class"  # classes should include both the class' and the __init__ method's docstring

autosummary_generate = True

todo_include_todos = True



import os
import json
import pytest
from network_wrangler import RoadwayNetwork
from network_wrangler import TransitNetwork
from network_wrangler import ProjectCard


"""
Run just the tests labeled transit using `pytest -v -m transit`
"""

STPAUL_DIR = os.path.join(os.getcwd(), "examples", "stpaul")
SCRATCH_DIR = os.path.join(os.getcwd(), "scratch")


@pytest.mark.roadway
@pytest.mark.transit
@pytest.mark.travis
def test_set_roadnet(request):
    print("\n--Starting:", request.node.name)

    road_net = RoadwayNetwork.read(
        link_file=os.path.join(STPAUL_DIR, "link.json"),
        node_file=os.path.join(STPAUL_DIR, "node.geojson"),
        shape_file=os.path.join(STPAUL_DIR, "shape.geojson"),
        fast=True,
    )
    transit_net = TransitNetwork.read(STPAUL_DIR)
    transit_net.set_roadnet(road_net)

    print("--Finished:", request.node.name)


@pytest.mark.roadway
@pytest.mark.transit
@pytest.mark.travis
# @pytest.mark.skip("")
def test_project_card(request):
    print("\n--Starting:", request.node.name)

    road_net = RoadwayNetwork.read(
        link_file=os.path.join(STPAUL_DIR, "link.json"),
        node_file=os.path.join(STPAUL_DIR, "node.geojson"),
        shape_file=os.path.join(STPAUL_DIR, "shape.geojson"),
        fast=True,
    )
    transit_net = TransitNetwork.read(STPAUL_DIR)
    transit_net.road_net = road_net
    project_card_path = os.path.join(
        STPAUL_DIR, "project_cards", "12_transit_shape_change.yml"
    )
    project_card = ProjectCard.read(project_card_path)
    transit_net.apply_transit_feature_change(
        transit_net.select_transit_features(project_card.facility),
        project_card.properties,
    )

    # Shapes
    result = transit_net.feed.shapes[transit_net.feed.shapes["shape_id"] == "2940002"][
        "shape_model_node_id"
    ].tolist()
    answer = [
        "37582",
        "37574",
        "4761",
        "4763",
        "4764",
        "98429",
        "45985",
        "57483",
        "126324",
        "57484",
        "150855",
        "11188",
        "84899",
        "46666",
        "46665",
        "46663",
        "81820",
        "76167",
        "77077",
        "68609",
        "39425",
        "62146",
        "41991",
        "70841",
        "45691",
        "69793",
        "45683",
        "45685",
        "7688",
        "45687",
        "100784",
        "100782",
        "45688",
        "37609",
        "19077",
        "38696",
    ]
    assert result == answer

    # Stops
    result = transit_net.feed.stop_times[
        transit_net.feed.stop_times["trip_id"] == "14944022-JUN19-MVS-BUS-Weekday-01"
    ]["stop_id"].tolist()
    result_tail = result[-5:]
    answer_tail = ["17013", "17010", "17009", "17006", "17005"]
    assert result_tail == answer_tail

    print("--Finished:", request.node.name)


@pytest.mark.roadway
@pytest.mark.transit
@pytest.mark.travis
# @pytest.mark.skip("need to allow for creating new stops if they don't already exist in stops.txt")
def test_wo_existing(request):
    print("\n--Starting:", request.node.name)

    road_net = RoadwayNetwork.read(
        link_file=os.path.join(STPAUL_DIR, "link.json"),
        node_file=os.path.join(STPAUL_DIR, "node.geojson"),
        shape_file=os.path.join(STPAUL_DIR, "shape.geojson"),
        fast=True,
    )
    transit_net = TransitNetwork.read(STPAUL_DIR)
    transit_net.road_net = road_net
    # A new node ID (not in stops.txt) should fail right now
    """with pytest.raises(Exception):
        transit_net.apply_transit_feature_change(
            trip_ids=transit_net.select_transit_features(
                {"trip_id": ["14944022-JUN19-MVS-BUS-Weekday-01"]}
            ),
            properties=[
                {
                    "property": "routing",
                    "set": [1]
                }
            ]
        )"""

    transit_net.apply_transit_feature_change(
        trip_ids=transit_net.select_transit_features(
            {"trip_id": ["14986385-JUN19-MVS-BUS-Weekday-01"]}
        ),
        properties=[{"property": "routing", "set": [75318]}],
    )

    # Shapes
    """result = transit_net.feed.shapes[
        transit_net.feed.shapes["shape_id"] == "210005"
    ]["shape_model_node_id"].tolist()
    answer = ["1"]
    assert result == answer"""

    # Stops
    result = transit_net.feed.stop_times[
        transit_net.feed.stop_times["trip_id"] == "14986385-JUN19-MVS-BUS-Weekday-01"
    ]["stop_id"].tolist()
    answer = ["2609"]  # first matching stop_id in stops.txt
    assert result == answer

    print("--Finished:", request.node.name)


@pytest.mark.roadway
@pytest.mark.transit
@pytest.mark.travis
# @pytest.mark.skip("need to update trips and nodes")
def test_select_transit_features_by_nodes(request):
    print("\n--Starting:", request.node.name)

    transit_net = TransitNetwork.read(STPAUL_DIR)

    # Any nodes
    trip_ids = transit_net.select_transit_features_by_nodes(
        node_ids=["75520", "66380", "57530"]
    )
    print(trip_ids)
    assert set(trip_ids) == set(
        [
            "14941148-JUN19-MVS-BUS-Weekday-01",
            "14941151-JUN19-MVS-BUS-Weekday-01",
            "14941153-JUN19-MVS-BUS-Weekday-01",
            "14941163-JUN19-MVS-BUS-Weekday-01",
            "14944379-JUN19-MVS-BUS-Weekday-01",
            "14944386-JUN19-MVS-BUS-Weekday-01",
            "14944413-JUN19-MVS-BUS-Weekday-01",
            "14944416-JUN19-MVS-BUS-Weekday-01",
        ]
    )

    # All nodes
    trip_ids = transit_net.select_transit_features_by_nodes(
        node_ids=["75520", "66380"], require_all=True
    )
    assert set(trip_ids) == set(
        [
            "14941148-JUN19-MVS-BUS-Weekday-01",
            "14941151-JUN19-MVS-BUS-Weekday-01",
            "14941153-JUN19-MVS-BUS-Weekday-01",
            "14941163-JUN19-MVS-BUS-Weekday-01",
        ]
    )

    print("--Finished:", request.node.name)


@pytest.mark.roadway
@pytest.mark.transit
@pytest.mark.travis
# @pytest.mark.skip("need to update trips and nodes")
def test_select_transit_features_by_nodes(request):
    print("\n--Starting:", request.node.name)

    transit_net = TransitNetwork.read(STPAUL_DIR)

    # Any nodes
    trip_ids = transit_net.select_transit_features_by_nodes(
        node_ids=["75520", "66380", "57530"]
    )
    print(trip_ids)
    assert set(trip_ids) == set(
        [
            "14941148-JUN19-MVS-BUS-Weekday-01",
            "14941151-JUN19-MVS-BUS-Weekday-01",
            "14941153-JUN19-MVS-BUS-Weekday-01",
            "14941163-JUN19-MVS-BUS-Weekday-01",
            "14944379-JUN19-MVS-BUS-Weekday-01",
            "14944386-JUN19-MVS-BUS-Weekday-01",
            "14944413-JUN19-MVS-BUS-Weekday-01",
            "14944416-JUN19-MVS-BUS-Weekday-01",
        ]
    )

    # All nodes
    trip_ids = transit_net.select_transit_features_by_nodes(
        node_ids=["75520", "66380"], require_all=True
    )
    assert set(trip_ids) == set(
        [
            "14941148-JUN19-MVS-BUS-Weekday-01",
            "14941151-JUN19-MVS-BUS-Weekday-01",
            "14941153-JUN19-MVS-BUS-Weekday-01",
            "14941163-JUN19-MVS-BUS-Weekday-01",
        ]
    )

    print("--Finished:", request.node.name)



import os
import json
import pytest
from network_wrangler import TransitNetwork
from network_wrangler import ProjectCard
from network_wrangler import RoadwayNetwork

"""
Run just the tests labeled transit using `pytest -v -m transit`
"""

STPAUL_DIR = os.path.join(os.getcwd(), "examples", "stpaul")
SCRATCH_DIR = os.path.join(os.getcwd(), "scratch")


@pytest.mark.basic
@pytest.mark.travis
@pytest.mark.transit
def test_transit_read_write(request):
    print("\n--Starting:", request.node.name)
    transit_net = TransitNetwork.read(feed_path=STPAUL_DIR)
    print("Transit Net Directory:", STPAUL_DIR)

    transit_net.write(path=SCRATCH_DIR)
    print("Transit Write Directory:", SCRATCH_DIR)

    print("--Finished:", request.node.name)


@pytest.mark.travis
@pytest.mark.transit
def test_select_transit_features(request):
    print("\n--Starting:", request.node.name)
    net = TransitNetwork.read(STPAUL_DIR)

    test_selections = {
        "1. simple trip_id": {
            "trip_id": "14940701-JUN19-MVS-BUS-Weekday-01",
            "answer": ["14940701-JUN19-MVS-BUS-Weekday-01"],
        },
        "2. multiple trip_id": {
            "trip_id": [
                "14969841-JUN19-RAIL-Weekday-01",  # unordered
                "14940701-JUN19-MVS-BUS-Weekday-01",
            ],
            "answer": [
                "14940701-JUN19-MVS-BUS-Weekday-01",
                "14969841-JUN19-RAIL-Weekday-01",
            ],
        },
        "3. route_id": {
            "route_id": "365-111",
            "answer": ["14947182-JUN19-MVS-BUS-Weekday-01"],
        },
    }

    for i, sel in test_selections.items():
        print("--->", i, "\n", sel)
        answer = sel.pop("answer")
        selected_trips = net.select_transit_features(sel)
        assert set(selected_trips) == set(answer)

    print("--Finished:", request.node.name)


@pytest.mark.transit
@pytest.mark.travis
def test_select_transit_features_from_projectcard(request):
    print("\n--Starting:", request.node.name)
    net = TransitNetwork.read(STPAUL_DIR)

    test_selections = [
        {
            "file": "7_simple_transit_attribute_change.yml",
            "answer": ["14940701-JUN19-MVS-BUS-Weekday-01"],
        },
        {
            "file": "7a_multi_transit_attribute_change.yml",
            "answer": [
                "14940701-JUN19-MVS-BUS-Weekday-01",
                "14948032-JUN19-MVS-BUS-Weekday-01",
            ],
        },
        {
            "file": "8_simple_transit_attribute_change.yml",
            "answer": [
                "14944012-JUN19-MVS-BUS-Weekday-01",
                "14944018-JUN19-MVS-BUS-Weekday-01",
                "14944019-JUN19-MVS-BUS-Weekday-01",
                "14944022-JUN19-MVS-BUS-Weekday-01",
            ],
        },
        {
            "file": "8a_multi_transit_attribute_change.yml",
            "answer": [
                "14944012-JUN19-MVS-BUS-Weekday-01",
                "14944018-JUN19-MVS-BUS-Weekday-01",
                "14944019-JUN19-MVS-BUS-Weekday-01",
                "14944022-JUN19-MVS-BUS-Weekday-01",
                "14948211-JUN19-MVS-BUS-Weekday-01",  # additional for 53-111
                "14948218-JUN19-MVS-BUS-Weekday-01",
            ],
        },
        {
            "file": "9_simple_transit_attribute_change.yml",
            "answer": [
                "14940701-JUN19-MVS-BUS-Weekday-01",
                "14943414-JUN19-MVS-BUS-Weekday-01",
                "14943415-JUN19-MVS-BUS-Weekday-01",
                "14946111-JUN19-MVS-BUS-Weekday-01",
                "14946257-JUN19-MVS-BUS-Weekday-01",
                "14946470-JUN19-MVS-BUS-Weekday-01",
                "14946471-JUN19-MVS-BUS-Weekday-01",
                "14946480-JUN19-MVS-BUS-Weekday-01",
                "14946521-JUN19-MVS-BUS-Weekday-01",
                "14947182-JUN19-MVS-BUS-Weekday-01",
                "14947504-JUN19-MVS-BUS-Weekday-01",
                "14947734-JUN19-MVS-BUS-Weekday-01",
                "14947755-JUN19-MVS-BUS-Weekday-01",
                "14978409-JUN19-MVS-BUS-Weekday-01",
                "14981028-JUN19-MVS-BUS-Weekday-01",
                "14981029-JUN19-MVS-BUS-Weekday-01",
                "14986383-JUN19-MVS-BUS-Weekday-01",
                "14986385-JUN19-MVS-BUS-Weekday-01",
            ],
        },
        {
            "file": "9a_multi_transit_attribute_change.yml",
            "answer": [
                "14940701-JUN19-MVS-BUS-Weekday-01",
                "14943414-JUN19-MVS-BUS-Weekday-01",
                "14943415-JUN19-MVS-BUS-Weekday-01",
                "14946111-JUN19-MVS-BUS-Weekday-01",
                "14946257-JUN19-MVS-BUS-Weekday-01",
                "14946470-JUN19-MVS-BUS-Weekday-01",
                "14946471-JUN19-MVS-BUS-Weekday-01",
                "14946480-JUN19-MVS-BUS-Weekday-01",
                "14946521-JUN19-MVS-BUS-Weekday-01",
                "14947182-JUN19-MVS-BUS-Weekday-01",
                "14947504-JUN19-MVS-BUS-Weekday-01",
                "14947734-JUN19-MVS-BUS-Weekday-01",
                "14947755-JUN19-MVS-BUS-Weekday-01",
                "14978409-JUN19-MVS-BUS-Weekday-01",
                "14981028-JUN19-MVS-BUS-Weekday-01",
                "14981029-JUN19-MVS-BUS-Weekday-01",
                "14986383-JUN19-MVS-BUS-Weekday-01",
                "14986385-JUN19-MVS-BUS-Weekday-01",
                "14946199-JUN19-MVS-BUS-Weekday-01",  # add below for Ltd Stop
                "14947598-JUN19-MVS-BUS-Weekday-01",
                "14948211-JUN19-MVS-BUS-Weekday-01",
                "14948218-JUN19-MVS-BUS-Weekday-01",
            ],
        },
    ]

    for i, test in enumerate(test_selections):
        print("--->", i)
        print("Reading project card", test["file"], "...")

        project_card_path = os.path.join(STPAUL_DIR, "project_cards", test["file"])
        project_card = ProjectCard.read(project_card_path)
        sel = project_card.facility

        selected_trips = net.select_transit_features(sel)
        assert set(selected_trips) == set(test["answer"])

    print("--Finished:", request.node.name)


@pytest.mark.transit
@pytest.mark.travis
def test_apply_transit_feature_change_from_projectcard(request):
    print("\n--Starting:", request.node.name)
    net = TransitNetwork.read(STPAUL_DIR)

    test_selections = [
        {
            "file": "7_simple_transit_attribute_change.yml",
            "answer": {
                "trip_ids": ["14940701-JUN19-MVS-BUS-Weekday-01"],
                "headway_secs": [900],
            },
        },
        {
            "file": "8_simple_transit_attribute_change.yml",
            "answer": {
                "trip_ids": [
                    "14944012-JUN19-MVS-BUS-Weekday-01",
                    "14944019-JUN19-MVS-BUS-Weekday-01",
                ],
                "headway_secs": [253, 226],
            },
        },
        {
            "file": "9_simple_transit_attribute_change.yml",
            "answer": {
                "trip_ids": [
                    "14940701-JUN19-MVS-BUS-Weekday-01",
                    "14943414-JUN19-MVS-BUS-Weekday-01",
                    "14943415-JUN19-MVS-BUS-Weekday-01",
                    "14946111-JUN19-MVS-BUS-Weekday-01",
                    "14946257-JUN19-MVS-BUS-Weekday-01",
                    "14946470-JUN19-MVS-BUS-Weekday-01",
                    "14946471-JUN19-MVS-BUS-Weekday-01",
                    "14946480-JUN19-MVS-BUS-Weekday-01",
                    "14946521-JUN19-MVS-BUS-Weekday-01",
                    "14947182-JUN19-MVS-BUS-Weekday-01",
                    "14947504-JUN19-MVS-BUS-Weekday-01",
                    "14947734-JUN19-MVS-BUS-Weekday-01",
                    "14947755-JUN19-MVS-BUS-Weekday-01",
                    "14978409-JUN19-MVS-BUS-Weekday-01",
                    "14981028-JUN19-MVS-BUS-Weekday-01",
                    "14981029-JUN19-MVS-BUS-Weekday-01",
                    "14986383-JUN19-MVS-BUS-Weekday-01",
                    "14986385-JUN19-MVS-BUS-Weekday-01",
                ],
                "headway_secs": [1800],
            },
        },
    ]

    for i, test in enumerate(test_selections):
        print("--->", i)
        print("Reading project card", test["file"], "...")

        project_card_path = os.path.join(STPAUL_DIR, "project_cards", test["file"])
        project_card = ProjectCard.read(project_card_path)
        net.apply_transit_feature_change(
            net.select_transit_features(project_card.facility), project_card.properties
        )

        freq = net.feed.frequencies
        answers = test["answer"]["headway_secs"]
        if len(answers) > 1:
            for i, answer in enumerate(answers):
                match = freq.trip_id == test["answer"]["trip_ids"][i]
                result = freq[match]["headway_secs"]
                assert set(result) == set([answer])
        else:
            matches = freq.trip_id.isin(test["answer"]["trip_ids"])
            results = freq[matches]["headway_secs"].tolist()
            assert set(results) == set(answers)

    print("--Finished:", request.node.name)


@pytest.mark.transit
@pytest.mark.travis
def test_wrong_existing(request):
    print("\n--Starting:", request.node.name)
    net = TransitNetwork.read(STPAUL_DIR)

    selected_trips = net.select_transit_features(
        {
            "trip_id": [
                "14944018-JUN19-MVS-BUS-Weekday-01",
                "14944012-JUN19-MVS-BUS-Weekday-01",
            ]
        }
    )

    with pytest.raises(Exception):
        net.apply_transit_feature_change(
            selected_trips, [{"property": "headway_secs", "existing": 553, "set": 900}]
        )

    print("--Finished:", request.node.name)


@pytest.mark.transit
@pytest.mark.travis
def test_zero_valid_facilities(request):
    print("\n--Starting:", request.node.name)
    net = TransitNetwork.read(STPAUL_DIR)

    with pytest.raises(Exception):
        net.select_transit_features(
            {
                "trip_id": ["14941433-JUN19-MVS-BUS-Weekday-01"],
                "time": ["06:00:00", "09:00:00"],
            }
        )

    print("--Finished:", request.node.name)


@pytest.mark.transit
@pytest.mark.travis
def test_invalid_selection_key(request):
    print("\n--Starting:", request.node.name)
    net = TransitNetwork.read(STPAUL_DIR)

    with pytest.raises(Exception):
        # trip_ids rather than trip_id should fail
        net.select_transit_features({"trip_ids": ["14941433-JUN19-MVS-BUS-Weekday-01"]})

    print("--Finished:", request.node.name)


@pytest.mark.transit
@pytest.mark.travis
def test_invalid_optional_selection_variable(request):
    print("\n--Starting:", request.node.name)
    net = TransitNetwork.read(STPAUL_DIR)

    with pytest.raises(Exception):
        # `wheelchair` rather than `wheelchair_accessible`
        net.select_transit_features(
            {"trip_id": "14940701-JUN19-MVS-BUS-Weekday-01", "wheelchair": "0"}
        )

    with pytest.raises(Exception):
        # Missing trip_id, route_id, route_short_name, or route_long_name
        net.select_transit_features({"wheelchair_accessible": "0"})

    # Correct trip variable
    sel = net.select_transit_features(
        {"trip_id": "14940701-JUN19-MVS-BUS-Weekday-01", "wheelchair_accessible": "1"}
    )
    assert set(sel) == set(["14940701-JUN19-MVS-BUS-Weekday-01"])

    # Correct route variable
    sel = net.select_transit_features({"route_long_name": "Express", "agency_id": "2"})
    assert set(sel) == set(["14978409-JUN19-MVS-BUS-Weekday-01"])

    print("--Finished:", request.node.name)


@pytest.mark.travis
def test_transit_road_consistencies(request):
    print("\n--Starting:", request.node.name)
    net = TransitNetwork.read(STPAUL_DIR)

    STPAUL_SHAPE_FILE = os.path.join(STPAUL_DIR, "shape.geojson")
    STPAUL_LINK_FILE = os.path.join(STPAUL_DIR, "link.json")
    STPAUL_NODE_FILE = os.path.join(STPAUL_DIR, "node.geojson")

    road_net = RoadwayNetwork.read(
        link_file=STPAUL_LINK_FILE,
        node_file=STPAUL_NODE_FILE,
        shape_file=STPAUL_SHAPE_FILE,
        fast=True,
    )

    net.set_roadnet(road_net=road_net)

    net.validate_road_network_consistencies()
    print(net.validated_road_network_consistency)
    print("--Finished:", request.node.name)


if __name__ == "__main__":
    test_transit_read_write()



import os
import pytest


@pytest.mark.travis
@pytest.mark.dependencies
def test_dependencies_api(request):
    print("\n--Starting:", request.node.name)
    from jsonschema import validate
    from jsonschema.exceptions import ValidationError
    from jsonschema.exceptions import SchemaError
    from folium import Icon, Circle, Marker
    from networkx import (
        shortest_path,
        NetworkXNoPath,
        is_strongly_connected,
        strongly_connected_components,
        DiGraph,
        MultiDiGraph,
    )
    from numpy import float64, int64, arange
    from osmnx import graph_from_gdfs, plot_graph, plot_graph_folium
    from osmnx.folium import _make_folium_polyline
    from geopandas.geodataframe import GeoDataFrame
    from pandas.core.frame import DataFrame
    from shapely.geometry import Point, LineString
    from partridge import load_feed
    from partridge.config import default_config



import os
import copy
import hashlib
import math
from typing import Union

from shapely.geometry import LineString

import pandas as pd
import geopandas as gpd
from shapely.geometry import LineString
from geographiclib.geodesic import Geodesic

from .logger import WranglerLogger


def point_df_to_geojson(df: pd.DataFrame, properties: list):
    """
    Author: Geoff Boeing:
    https://geoffboeing.com/2015/10/exporting-python-data-geojson/
    """
    from .roadwaynetwork import RoadwayNetwork

    geojson = {"type": "FeatureCollection", "features": []}
    for _, row in df.iterrows():
        feature = {
            "type": "Feature",
            "properties": {},
            "geometry": {"type": "Point", "coordinates": []},
        }
        feature["geometry"]["coordinates"] = [row["geometry"].x, row["geometry"].y]
        feature["properties"][RoadwayNetwork.NODE_FOREIGN_KEY] = row.name
        for prop in properties:
            feature["properties"][prop] = row[prop]
        geojson["features"].append(feature)
    return geojson


def link_df_to_json(df: pd.DataFrame, properties: list):
    """ Export pandas dataframe as a json object.

    Modified from: Geoff Boeing:
    https://geoffboeing.com/2015/10/exporting-python-data-geojson/

    Args:
        df: Dataframe to export
        properties: list of properties to export
    """

    # can't remember why we need this?
    if "distance" in properties:
        df["distance"].fillna(0)

    json = []
    for _, row in df.iterrows():
        feature = {}
        for prop in properties:
            feature[prop] = row[prop]
        json.append(feature)

    return json


def topological_sort(adjacency_list, visited_list):
    """
    Topological sorting for Acyclic Directed Graph
    """

    output_stack = []

    def _topology_sort_util(vertex):
        if not visited_list[vertex]:
            visited_list[vertex] = True
            for neighbor in adjacency_list[vertex]:
                _topology_sort_util(neighbor)
            output_stack.insert(0, vertex)

    for vertex in visited_list:
        _topology_sort_util(vertex)

    return output_stack


def make_slug(text, delimiter: str = "_"):
    """
    makes a slug from text
    """
    import re

    text = re.sub("[,.;@#?!&$']+", "", text.lower())
    return re.sub("[\ ]+", delimiter, text)


def parse_time_spans(times):
    """
    parse time spans into tuples of seconds from midnight
    can also be used as an apply function for a pandas series
    Parameters
    -----------
    times: tuple(string) or tuple(int) or list(string) or list(int)

    returns
    --------
    tuple(integer)
      time span as seconds from midnight
    """
    try:
        start_time, end_time = times
    except:
        msg = "ERROR: times should be a tuple or list of two, got: {}".format(times)
        WranglerLogger.error(msg)
        raise ValueError(msg)

    # If times are strings, convert to int in seconds, else return as ints
    if isinstance(start_time, str) and isinstance(end_time, str):
        start_time = start_time.strip()
        end_time = end_time.strip()

        # If time is given without seconds, add 00
        if len(start_time) <= 5:
            start_time += ":00"
        if len(end_time) <= 5:
            end_time += ":00"

        # Convert times to seconds from midnight (Partride's time storage)
        h0, m0, s0 = start_time.split(":")
        start_time_sec = int(h0) * 3600 + int(m0) * 60 + int(s0)

        h1, m1, s1 = end_time.split(":")
        end_time_sec = int(h1) * 3600 + int(m1) * 60 + int(s1)

        return (start_time_sec, end_time_sec)

    elif isinstance(start_time, int) and isinstance(end_time, int):
        return times

    else:
        WranglerLogger.error("ERROR: times should be ints or strings")
        raise ValueError()

    return (start_time_sec, end_time_sec)


def get_bearing(lat1, lon1, lat2, lon2):
    """
    calculate the bearing (forward azimuth) b/w the two points

    returns: bearing in radians
    """
    # bearing in degrees
    brng = Geodesic.WGS84.Inverse(lat1, lon1, lat2, lon2)["azi1"]

    # convert bearing to radians
    brng = math.radians(brng)

    return brng


def offset_point_with_distance_and_bearing(lat, lon, distance, bearing):
    """
    Get the new lat long (in degrees) given current point (lat/lon), distance and bearing

    returns: new lat/long
    """
    # Earth's radius in meters
    radius = 6378137

    # convert the lat long from degree to radians
    lat_radians = math.radians(lat)
    lon_radians = math.radians(lon)

    # calculate the new lat long in radians
    out_lat_radians = math.asin(
        math.sin(lat_radians) * math.cos(distance / radius)
        + math.cos(lat_radians) * math.sin(distance / radius) * math.cos(bearing)
    )

    out_lon_radians = lon_radians + math.atan2(
        math.sin(bearing) * math.sin(distance / radius) * math.cos(lat_radians),
        math.cos(distance / radius) - math.sin(lat_radians) * math.sin(lat_radians),
    )
    # convert the new lat long back to degree
    out_lat = math.degrees(out_lat_radians)
    out_lon = math.degrees(out_lon_radians)

    return (out_lat, out_lon)


def offset_location_reference(location_reference, offset_meters=10):
    """
    Creates a new location reference
    using the node a and node b of given location reference,
    offseting it by 90 degree to the bearing of given location reference
    and distance equals to offset_meters

    returns: new location_reference with offset
    """
    lon_1 = location_reference[0]["point"][0]
    lat_1 = location_reference[0]["point"][1]
    lon_2 = location_reference[1]["point"][0]
    lat_2 = location_reference[1]["point"][1]

    bearing = get_bearing(lat_1, lon_1, lat_2, lon_2)
    # adding 90 degrees (1.57 radians) to the current bearing
    bearing = bearing + 1.57

    new_lat_1, new_lon_1 = offset_point_with_distance_and_bearing(
        lat_1, lon_1, offset_meters, bearing
    )
    new_lat_2, new_lon_2 = offset_point_with_distance_and_bearing(
        lat_2, lon_2, offset_meters, bearing
    )

    out_location_reference = [
        {"sequence": 1, "point": [new_lon_1, new_lat_1]},
        {"sequence": 2, "point": [new_lon_2, new_lat_2]},
    ]

    return out_location_reference


def haversine_distance(origin: list, destination: list):
    """
    Calculates haversine distance between two points

    Args:
    origin: lat/lon for point A
    destination: lat/lon for point B

    Returns: string
    """

    lon1, lat1 = origin
    lon2, lat2 = destination
    radius = 6378137  # meter

    dlat = math.radians(lat2 - lat1)
    dlon = math.radians(lon2 - lon1)
    a = math.sin(dlat / 2) * math.sin(dlat / 2) + math.cos(
        math.radians(lat1)
    ) * math.cos(math.radians(lat2)) * math.sin(dlon / 2) * math.sin(dlon / 2)
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
    d = radius * c  # meters
    d = d * 0.000621371  # miles

    return d


def create_unique_shape_id(line_string: LineString):
    """
    Creates a unique hash id using the coordinates of the geomtery

    Args:
    line_string: Line Geometry as a LineString

    Returns: string
    """

    x1, y1 = list(line_string.coords)[0]  # first co-ordinate (A node)
    x2, y2 = list(line_string.coords)[-1]  # last co-ordinate (B node)

    message = "Geometry {} {} {} {}".format(x1, y1, x2, y2)
    unhashed = message.encode("utf-8")
    hash = hashlib.md5(unhashed).hexdigest()

    return hash


def create_location_reference_from_nodes(node_a, node_b):
    """
    Creates a location reference using the node a and node b coordinates

    Args:
    node_a: Node A as Series
    node_b: Node B as Series

    """
    out_location_reference = [
        {"sequence": 1, "point": [node_a["X"], node_a["Y"]]},
        {"sequence": 2, "point": [node_b["X"], node_b["Y"]]},
    ]

    return out_location_reference


def create_line_string(location_reference: list):
    """
    Creates a geometry as a LineString using location reference
    """

    return LineString([location_reference[0]["point"], location_reference[1]["point"]])



import os
import pytest

from shapely.geometry import LineString

from network_wrangler import haversine_distance
from network_wrangler import create_unique_shape_id
from network_wrangler import offset_location_reference

slug_test_list = [
    {"text": "I am a roadway", "delim": "_", "answer": "i_am_a_roadway"},
    {"text": "I'm a roadway", "delim": "_", "answer": "im_a_roadway"},
    {"text": "I am a roadway", "delim": "-", "answer": "i-am-a-roadway"},
    {"text": "I am a roadway", "delim": "", "answer": "iamaroadway"},
]


@pytest.mark.travis
@pytest.mark.parametrize("slug_test", slug_test_list)
def test_get_slug(request, slug_test):
    print("\n--Starting:", request.node.name)

    from network_wrangler.utils import make_slug

    slug = make_slug(slug_test["text"], delimiter=slug_test["delim"])

    print("From: {} \nTo: {}".format(slug_test["text"], slug))
    print("Expected: {}".format(slug_test["answer"]))
    assert slug == slug_test["answer"]


@pytest.mark.travis
def test_time_convert(request):
    print("\n--Starting:", request.node.name)

    time_tests = [
        (("00:00:00", "00:00:10"), (0, 10)),
        (("0:00", "0:10:00"), (0, 600)),
        (("01:02:03", "01:02:23"), (3723, 3743)),
        (("1:02", "1:02:13"), (3720, 3733)),
        (("25:24:23", "25:24:33"), (91463, 91473)),
        (("250:24:23", "250:24:25"), (901463, 901465)),
    ]

    from pandas import DataFrame

    df = DataFrame(time_tests, columns=["time", "time_results"])
    print("Original Time Series", df)

    from network_wrangler.utils import parse_time_spans

    df["time"] = df["time"].apply(parse_time_spans)
    print("Result Time Series", df)
    from pandas.testing import assert_series_equal

    assert_series_equal(df["time"], df["time_results"], check_names=False)


@pytest.mark.get_dist
@pytest.mark.travis
def test_get_distance_bw_lat_lon(request):
    print("\n--Starting:", request.node.name)

    start = [-93.0889873, 44.966861]
    end = [-93.08844310000001, 44.9717832]
    dist = haversine_distance(start, end)
    print(dist)
    assert dist == 0.34151200885686445
    print("--Finished:", request.node.name)


@pytest.mark.test_hash
@pytest.mark.roadway
def test_get_unique_shape_id(request):
    geometry = LineString([[-93.0855338, 44.9662078], [-93.0843092, 44.9656997]])

    shape_id = create_unique_shape_id(geometry)

    assert shape_id == "72ceb24e2c632c02f7eae5e33ed12702"

    print("--Finished:", request.node.name)


@pytest.mark.travis
def test_location_reference_offset(request):
    print("\n--Starting:", request.node.name)

    location_reference = [
        {"sequence": 1, "point": [-93.0903549, 44.961085]},
        {"sequence": 2, "point": [-93.0889873, 44.966861]},
    ]

    print("original ref", location_reference)

    expected_location_reference = [
        {"sequence": 1, "point": [-93.09022968479499, 44.961070179988084]},
        {"sequence": 2, "point": [-93.08886207218725, 44.966846179988075]},
    ]

    new_location_reference = offset_location_reference(location_reference)
    print("new ref", new_location_reference)

    assert new_location_reference == expected_location_reference

    print("--Finished:", request.node.name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Utility functions for logging

import sys
import logging

__all__ = ["WranglerLogger", "setupLogging"]

WranglerLogger = logging.getLogger("WranglerLogger")


def setupLogging(
    level: int = None, log_filename: str = None, log_to_console: bool = False
):
    """
    Sets up the WranglerLogger w.r.t. the debug file location and if logging to console.

    args:
        level: the level of logging that will be recorded
        log_filename: the location of the log file that will get created to add the DEBUG log
        log_to_console: if True, logging will go to the console at INFO level
        """

    if level is None:
        WranglerLogger.setLevel(logging.DEBUG)
    else:
        WranglerLogger.setLevel(level)

    FORMAT = logging.Formatter(
        "%(asctime)-15s %(levelname)s: %(message)s", datefmt="%Y-%m-%d %H:%M:%S,"
    )

    if log_filename:
        file_handler = logging.FileHandler(log_filename)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(FORMAT)
        WranglerLogger.addHandler(file_handler)

    if log_to_console:
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setLevel(logging.INFO)
        console_handler.setFormatter(FORMAT)
        WranglerLogger.addHandler(console_handler)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations
import os
import sys
import glob
import copy
import pprint
from datetime import datetime
from typing import Union

import pandas as pd
import geopandas as gpd

from .projectcard import ProjectCard
from collections import OrderedDict
from .logger import WranglerLogger
from collections import defaultdict
from .utils import topological_sort
from .roadwaynetwork import RoadwayNetwork
from .transitnetwork import TransitNetwork


class Scenario(object):
    """
    Holds information about a scenario.

    .. highlight:: python

    Typical usage example:
    ::
        my_base_scenario = {
            "road_net": RoadwayNetwork.read(
                link_file=STPAUL_LINK_FILE,
                node_file=STPAUL_NODE_FILE,
                shape_file=STPAUL_SHAPE_FILE,
                fast=True,
            ),
            "transit_net": TransitNetwork.read(STPAUL_DIR),
        }

        card_filenames = [
            "3_multiple_roadway_attribute_change.yml",
            "multiple_changes.yml",
            "4_simple_managed_lane.yml",
        ]

        project_card_directory = os.path.join(STPAUL_DIR, "project_cards")

        project_cards_list = [
            ProjectCard.read(os.path.join(project_card_directory, filename), validate=False)
            for filename in card_filenames
        ]

        my_scenario = Scenario.create_scenario(
          base_scenario=my_base_scenario,
          project_cards_list=project_cards_list,
        )
        my_scenario.check_scenario_requisites()

        my_scenario.apply_all_projects()

        my_scenario.scenario_summary()

    Attributes:
        base_scenario: dictionary representation of a scenario
        project_cards (Optional): list of Project Card Instances
        road_net: instance of RoadwayNetwork for the scenario
        transit_net: instance of TransitNetwork for the scenario
        applied_projects: list of project names that have been applied
        project_cards: list of project card instances
        ordered_project_cards:
        prerequisites:  dictionary storing prerequiste information
        corequisites:  dictionary storing corequisite information
        conflicts: dictionary storing conflict information
        requisites_checked: boolean indicating if the co- and pre-requisites
            have been checked in the project cards
        conflicts_checked: boolean indicating if the project conflicts have been checked
        has_requisite_error: boolean indicating if there is a conflict in the pre- or
            co-requisites of project cards
        has_conflict_error: boolean indicating if there is are conflicting project cards
        prerequisites_sorted: boolean indicating if the project cards have
            been sorted to make sure cards that are pre-requisites are applied first
    """

    def __init__(self, base_scenario: dict, project_cards: [ProjectCard] = None):
        """
        Constructor

        args:
        base_scenario: dict the base scenario
        project_cards: list this scenario's project cards
        """

        self.road_net = None
        self.transit_net = None

        self.base_scenario = base_scenario

        # if the base scenario had roadway or transit networks, use them as the basis.
        if self.base_scenario.get("road_net"):
            self.road_net = copy.deepcopy(self.base_scenario["road_net"])
        if self.base_scenario.get("transit_net"):
            self.transit_net = copy.deepcopy(self.base_scenario["transit_net"])

        # if the base scenario had applied projects, add them to the list of applied
        self.applied_projects = []
        if self.base_scenario.get("applied_projects"):
            self.applied_projects = base_scenario["applied_projects"]

        self.project_cards = project_cards
        self.ordered_project_cards = OrderedDict()

        self.prerequisites = {}
        self.corequisites = {}
        self.conflicts = {}

        self.requisites_checked = False
        self.conflicts_checked = False

        self.has_requisite_error = False
        self.has_conflict_error = False

        self.prerequisites_sorted = False

        for card in self.project_cards:
            if not card.__dict__.get("dependencies"):
                continue

            if card.dependencies.get("prerequisites"):
                self.prerequisites.update(
                    {card.project: card.dependencies["prerequisites"]}
                )
            if card.dependencies.get("corequisites"):
                self.corequisites.update(
                    {card.project: card.dependencies["corequisites"]}
                )

    @staticmethod
    def create_base_scenario(
        base_shape_name: str,
        base_link_name: str,
        base_node_name: str,
        roadway_dir: str = "",
        transit_dir: str = "",
        validate: bool = True,
    ) -> Scenario:
        """
        args
        -----
        roadway_dir: optional
          path to the base scenario roadway network files
        base_shape_name:
          filename of the base network shape
        base_link_name:
          filename of the base network link
        base_node_name:
          filename of the base network node
        transit_dir: optional
           path to base scenario transit files
        validate:
          boolean indicating whether to validate the base network or not
        """
        if roadway_dir:
            base_network_shape_file = os.path.join(roadway_dir, base_shape_name)
            base_network_link_file = os.path.join(roadway_dir, base_link_name)
            base_network_node_file = os.path.join(roadway_dir, base_node_name)
        else:
            base_network_shape_file = base_shape_name
            base_network_link_file = base_link_name
            base_network_node_file = base_node_name

        road_net = RoadwayNetwork.read(
            link_file=base_network_link_file,
            node_file=base_network_node_file,
            shape_file=base_network_shape_file,
            fast=not validate,
        )

        if transit_dir:
            transit_net = TransitNetwork.read(transit_dir)
        else:
            transit_net = None
            WranglerLogger.info(
                "No transit directory specified, base scenario will have empty transit network."
            )

        transit_net.set_roadnet(road_net, validate_consistency=validate)
        base_scenario = {"road_net": road_net, "transit_net": transit_net}

        return base_scenario

    @staticmethod
    def create_scenario(
        base_scenario: dict = {},
        card_directory: str = "",
        tags: [str] = None,
        project_cards_list=[],
        glob_search="",
        validate_project_cards=True,
    ) -> Scenario:
        """
        Validates project cards with a specific tag from the specified folder or
        list of user specified project cards and
        creates a scenario object with the valid project card.

        args
        -----
        base_scenario:
          object dictionary for the base scenario (i.e. my_base_scenario.__dict__)
        tags:
          only project cards with these tags will be read/validated
        folder:
          the folder location where the project cards will be
        project_cards_list:
          list of project cards to be applied
        glob_search:

        """
        WranglerLogger.info("Creating Scenario")

        if project_cards_list:
            WranglerLogger.debug(
                "Adding project cards from List.\n{}".format(
                    ",".join([p.project for p in project_cards_list])
                )
            )
        scenario = Scenario(base_scenario, project_cards=project_cards_list)

        if card_directory and tags:
            WranglerLogger.debug(
                "Adding project cards from directory and tags.\nDir: {}\nTags: {}".format(
                    card_directory, ",".join(tags)
                )
            )
            scenario.add_project_cards_from_tags(
                card_directory,
                tags=tags,
                glob_search=glob_search,
                validate=validate_project_cards,
            )
        elif card_directory:
            WranglerLogger.debug(
                "Adding project cards from directory.\nDir: {}".format(card_directory)
            )
            scenario.add_project_cards_from_directory(
                card_directory, glob_search=glob_search, validate=validate_project_cards
            )
        return scenario

    def add_project_card_from_file(
        self, project_card_file: str, validate: bool = True, tags: list = []
    ):

        WranglerLogger.debug(
            "Trying to add project card from file: {}".format(project_card_file)
        )
        project_card = ProjectCard.read(project_card_file, validate=validate)

        if project_card == None:
            msg = "project card not read: {}".format(project_card_file)
            WranglerLogger.error(msg)
            raise ValueError(msg)

        if tags and set(tags).isdisjoint(project_card.tags):
            WranglerLogger.debug(
                "Project card tags: {} don't match search tags: {}".format(
                    ",".join(project_card.tags), ",".join(tags)
                )
            )
            return

        if project_card.project in self.get_project_names():
            msg = "project card with name '{}' already in Scenario. Project names must be unique".format(
                project_card.project
            )
            WranglerLogger.error(msg)
            raise ValueError(msg)

        self.requisites_checked = False
        self.conflicts_checked = False
        self.prerequisites_sorted = False

        WranglerLogger.debug(
            "Adding project card to scenario: {}".format(project_card.project)
        )
        self.project_cards.append(project_card)

        if not project_card.__dict__.get("dependencies"):
            return

        WranglerLogger.debug("Adding project card dependencies")
        if project_card.dependencies.get("prerequisites"):
            self.prerequisites.update(
                {project_card.project: project_card.dependencies["prerequisites"]}
            )
        if project_card.dependencies.get("corequisites"):
            self.corequisites.update(
                {project_card.project: project_card.dependencies["corequisites"]}
            )
        if project_card.dependencies.get("conflicts"):
            self.conflicts.update(
                {project_card.project: project_card.dependencies["conflicts"]}
            )

    def add_project_cards_from_directory(
        self, folder: str, glob_search="", validate=True
    ):
        """
        Adds projects cards to the scenario.
        A folder is provided to look for project cards and if applicable, a glob-style search.

        i.e. glob_search = 'road*.yml'

        args:
        folder: the folder location where the project cards will be
        glob_search: https://docs.python.org/2/library/glob.html
        """

        if not os.path.exists(folder):
            msg = "Cannot find specified directory to add project cards: {}".format(
                folder
            )
            WranglerLogger.error(msg)
            raise ValueError(msg)

        if glob_search:
            WranglerLogger.debug(
                "Adding project cards using glob search: {}".format(glob_search)
            )
            for file in glob.iglob(os.path.join(folder, glob_search)):
                if not file.endswith(".yml") or file.endswith(".yaml"):
                    continue
                else:
                    self.add_project_card_from_file(file, validate=validate)
        else:
            for file in os.listdir(folder):
                if not file.endswith(".yml") or file.endswith(".yaml"):
                    continue
                else:
                    self.add_project_card_from_file(
                        os.path.join(folder, file), validate=validate
                    )

    def add_project_cards_from_tags(
        self, folder: str, tags: [str] = [], glob_search="", validate=True
    ):
        """
        Adds projects cards to the scenario.
        A folder is provided to look for project cards that have a matching tag that is passed to the method.

        args:
        folder: the folder location where the project cards will be
        tags: only project cards with these tags will be validated and added to the returning scenario
        """

        if glob_search:
            WranglerLogger.debug(
                "Adding project cards using \n-tags: {} and \nglob search: {}".format(
                    tags, glob_search
                )
            )
            for file in glob.iglob(os.path.join(folder, glob_search)):
                self.add_project_card_from_file(file, tags=tags, validate=validate)
        else:
            WranglerLogger.debug("Adding project cards using \n-tags: {}".format(tags))
            for file in os.listdir(folder):

                self.add_project_card_from_file(file, tags=tags, validate=validate)

    def __str__(self):
        s = ["{}: {}".format(key, value) for key, value in self.__dict__.items()]
        return "\n".join(s)

    def get_project_names(self) -> list:
        """
        Returns a list of project names
        """
        return [project_card.project for project_card in self.project_cards]

    def check_scenario_conflicts(self) -> bool:
        """
        Checks if there are any conflicting projects in the scenario
        Fail if the project A specifies that project B is a conflict and project B is included in the scenario

        Returns: boolean indicating if the check was successful or returned an error
        """

        conflict_dict = self.conflicts
        scenario_projects = [p.project for p in self.project_cards]

        for project, conflicts in conflict_dict.items():
            if conflicts:
                for name in conflicts:
                    if name in scenario_projects:
                        self.project_cards
                        WranglerLogger.error(
                            "Projects %s has %s as conflicting project"
                            % (project, name)
                        )
                        self.has_conflict_error = True

        self.conflicts_checked = True

        return self.has_conflict_error

    def check_scenario_requisites(self) -> bool:
        """
        Checks if there are any missing pre- or co-requisite projects in the scenario
        Fail if the project A specifies that project B is a pre- or co-requisite and project B is not included in the scenario

        Returns: boolean indicating if the checks were successful or returned an error
        """

        corequisite_dict = self.corequisites
        prerequisite_dict = self.prerequisites

        scenario_projects = [p.project for p in self.project_cards]

        for project, coreq in corequisite_dict.items():
            if coreq:
                for name in coreq:
                    if name not in scenario_projects:
                        WranglerLogger.error(
                            "Projects %s has %s as corequisite project which is missing for the scenario"
                            % (project, name)
                        )
                        self.has_requisite_error = True

        for project, prereq in prerequisite_dict.items():
            if prereq:
                for name in prereq:
                    if name not in scenario_projects:
                        WranglerLogger.error(
                            "Projects %s has %s as prerequisite project which is missing for the scenario"
                            % (project, name)
                        )
                        self.has_requisite_error = True

        self.requisites_checked = True

        return self.has_requisite_error

    def order_project_cards(self):
        """
        create a list of project cards such that they are in order based on pre-requisites

        Returns: ordered list of project cards to be applied to scenario
        """

        scenario_projects = [p.project.lower() for p in self.project_cards]

        # build prereq (adjacency) list for topological sort
        adjacency_list = defaultdict(list)
        visited_list = defaultdict()

        for project in scenario_projects:
            visited_list[project] = False
            if not self.prerequisites.get(project):
                continue
            for prereq in self.prerequisites[project]:
                if (
                    prereq.lower() in scenario_projects
                ):  # this will always be true, else would have been flagged in missing prerequsite check, but just in case
                    adjacency_list[prereq.lower()] = [project]

        # sorted_project_names is topological sorted project card names (based on prerequsiite)
        sorted_project_names = topological_sort(
            adjacency_list=adjacency_list, visited_list=visited_list
        )

        # get the project card objects for these sorted project names
        project_card_and_name_dict = {}
        for project_card in self.project_cards:
            project_card_and_name_dict[project_card.project.lower()] = project_card

        sorted_project_cards = [
            project_card_and_name_dict[project_name]
            for project_name in sorted_project_names
        ]

        try:
            assert len(sorted_project_cards) == len(self.project_cards)
        except:
            msg = "Sorted project cards ({}) are not of same number as unsorted ({}).".format(
                len(sorted_project_cards), len(self.project_cards)
            )
            WranglerLogger.error(msg)
            raise ValueError(msg)

        self.prerequisites_sorted = True
        self.ordered_project_cards = {
            project_name: project_card_and_name_dict[project_name]
            for project_name in sorted_project_names
        }

        WranglerLogger.debug(
            "Ordered Project Cards: {}".format(self.ordered_project_cards)
        )
        self.project_cards = sorted_project_cards

        WranglerLogger.debug("Project Cards: {}".format(self.project_cards))

        return sorted_project_cards

    def apply_all_projects(self):

        # Get everything in order

        if not self.requisites_checked:
            self.check_scenario_requisites()
        if not self.conflicts_checked:
            self.check_scenario_conflicts()
        if not self.prerequisites_sorted:
            self.order_project_cards()

        for p in self.project_cards:
            self.apply_project(p.__dict__)

    def apply_project(self, p):
        if isinstance(p, ProjectCard):
            p = p.__dict__

        if p.get("project"):
            WranglerLogger.info("Applying {}".format(p["project"]))

        if p.get("changes"):
            part = 1
            for pc in p["changes"]:
                pc["project"] = p["project"]
                self.apply_project(pc)

        else:
            if p["category"] in ProjectCard.ROADWAY_CATEGORIES:
                if not self.road_net:
                    raise ("Missing Roadway Network")
                self.road_net.apply(p)
            if p["category"] in ProjectCard.TRANSIT_CATEGORIES:
                if not self.transit_net:
                    raise ("Missing Transit Network")
                self.transit_net.apply(p)
            if (
                p["category"] in ProjectCard.SECONDARY_TRANSIT_CATEGORIES
                and self.transit_net
            ):
                self.transit_net.apply(p)

            if p["project"] not in self.applied_projects:
                self.applied_projects.append(p["project"])

    def remove_all_projects(self):
        self.project_cards = []

    def applied_project_card_summary(self, project_card_dictionary: dict) -> dict:
        """
        Create a summary of applied project card and what they changed for the scenario.

        Args:
            project_card_dictionary: dictionary representation of the values of a project card (i.e. ProjectCard.__dict__ )

        Returns:
            A dict of project summary change dictionaries for each change
        """
        changes = project_card_dictionary.get("changes", [project_card_dictionary])

        summary = {
            "project_card": project_card_dictionary["file"],
            "total_parts": len(changes),
        }

        def _summarize_change_roadway(change: dict, change_summary: dict):

            sel_key = RoadwayNetwork.build_selection_key(
                self.road_net, change["facility"]
            )

            change_summary["sel_idx"] = self.road_net.selections[sel_key][
                "selected_links"
            ].index.tolist()

            change_summary["attributes"] = [p["property"] for p in change["properties"]]

            if type(sel_key) == tuple:
                _, A_id, B_id = sel_key
            else:
                A_id, B_id = (None, None)

            change_summary["map"] = self.road_net.selection_map(
                change_summary["sel_idx"],
                A=A_id,
                B=B_id,
                candidate_link_idx=self.road_net.selections[sel_key]
                .get("candidate_links", pd.DataFrame([]))
                .index.tolist(),
            )
            return change_summary

        def _summarize_add_roadway(change: dict, change_summary: dict):
            change_summary["added_links"] = pd.DataFrame(change.get("links"))
            change_summary["added_nodes"] = pd.DataFrame(change.get("nodes"))
            change_summary["map"] = RoadwayNetwork.addition_map(
                self.road_net, change.get("links"), change.get("nodes"),
            )
            return change_summary

        def _summarize_deletion(change: dict, change_summary: dict):
            change_summary["deleted_links"] = change.get("links")
            change_summary["deleted_nodes"] = change.get("nodes")
            change_summary["map"] = RoadwayNetwork.deletion_map(
                self.base_scenario["road_net"],
                change.get("links"),
                change.get("nodes"),
            )
            return change_summary

        for i, change in enumerate(changes):
            WranglerLogger.debug(
                "Summarizing {} Part: {}".format(
                    project_card_dictionary["project"], i + 1
                )
            )
            change_summary = {
                "project": project_card_dictionary["project"] + " – Part " + str(i + 1),
                "category": change["category"].lower(),
            }

            if change["category"].lower() == "roadway deletion":
                change_summary = _summarize_deletion(change, change_summary)
            elif change["category"].lower() == "add new roadway":
                change_summary = _summarize_add_roadway(change, change_summary)
            elif change["category"].lower() in [
                "roadway property change",
                "parallel managed lanes",
            ]:
                change_summary = _summarize_change_roadway(change, change_summary)

            summary["Part " + str(i + 1)] = change_summary

        return summary

    def scenario_summary(
        self, project_detail: bool = True, outfile: str = "", mode: str = "a"
    ) -> str:
        """
        A high level summary of the created scenario.

        Args:
            project_detail: If True (default), will write out project card summaries.
            outfile: If specified, will write scenario summary to text file.
            mode: Outfile open mode. 'a' to append 'w' to overwrite.

        Returns:
            string of summary

        """

        WranglerLogger.info("Summarizing Scenario")
        report_str = "------------------------------\n"
        report_str += "Scenario created on {}\n".format(datetime.now())

        report_str += "Base Scenario:\n"
        report_str += "--Road Network:\n"
        report_str += "----Link File: {}\n".format(
            self.base_scenario["road_net"].link_file
        )
        report_str += "----Node File: {}\n".format(
            self.base_scenario["road_net"].node_file
        )
        report_str += "----Shape File: {}\n".format(
            self.base_scenario["road_net"].shape_file
        )
        report_str += "--Transit Network:\n"
        report_str += "----Feed Path: {}\n".format(
            self.base_scenario["transit_net"].feed_path
        )

        report_str += "\nProject Cards:\n -"
        report_str += "\n-".join(p.file for p in self.project_cards)

        report_str += "\nApplied Projects:\n-"
        report_str += "\n-".join(self.applied_projects)

        if project_detail:
            report_str += "\n---Project Card Details---\n"
            for p in self.project_cards:
                report_str += "\n{}".format(
                    pprint.pformat(self.applied_project_card_summary(p.__dict__))
                )

        if outfile:
            with open(outfile, mode) as f:
                f.write(report_str)
            WranglerLogger.info("Wrote Scenario Report to: {}".format(outfile))

        return report_str


def net_to_mapbox(
    roadway: Union[RoadwayNetwork, gpd.GeoDataFrame] = gpd.GeoDataFrame(),
    transit: Union[TransitNetwork, gpd.GeoDataFrame] = gpd.GeoDataFrame(),
    roadway_geojson: str = "roadway_shapes.geojson",
    transit_geojson: str = "transit_shapes.geojson",
    mbtiles_out: str = "network.mbtiles",
    overwrite: bool = True,
    port: str = "9000",
):
    """

    Args:
        roadway: a RoadwayNetwork instance or a geodataframe with roadway linetrings
        transit: a TransitNetwork instance or a geodataframe with transit linetrings
    """
    import subprocess

    # test for mapbox token
    try:
        token = os.getenv("MAPBOX_ACCESS_TOKEN")
    except:
        raise (
            "NEED TO SET MAPBOX ACCESS TOKEN IN ENVIRONMENT VARIABLES/n In command line: >>export MAPBOX_ACCESS_TOKEN='pk.0000.1111' # replace value with your mapbox public access token"
        )

    if type(transit) != gpd.GeoDataFrame:
        transit = TransitNetwork.transit_net_to_gdf(transit)
    if type(roadway) != gpd.GeoDataFrame:
        roadway = RoadwayNetwork.roadway_net_to_gdf(roadway)

    transit.to_file(transit_geojson, driver="GeoJSON")
    roadway.to_file(roadway_geojson, driver="GeoJSON")

    tippe_options_list = ["-zg", "-o", mbtiles_out]
    if overwrite:
        tippe_options_list.append("--force")
    # tippe_options_list.append("--drop-densest-as-needed")
    tippe_options_list.append(roadway_geojson)
    tippe_options_list.append(transit_geojson)

    try:
        WranglerLogger.info(
            "Running tippecanoe with following options: {}".format(
                " ".join(tippe_options_list)
            )
        )
        tippe_out = subprocess.run(["tippecanoe"] + tippe_options_list)
    except:
        WranglerLogger.error()
        raise (
            "If tippecanoe isn't installed, try `brew install tippecanoe` or visit https://github.com/mapbox/tippecanoe"
        )

    try:
        WranglerLogger.info(
            "Running mbview with following options: {}".format(
                " ".join(tippe_options_list)
            )
        )
        mbview_out = subprocess.run(
            ["mbview", "--port", port, ",/{}".format(mbtiles_out)]
        )
    except:
        WranglerLogger.error()
        raise (
            "If mbview isn't installed, try `npm install -g @mapbox/mbview` or visit https://github.com/mapbox/mbview"
        )



import os
import sys
import subprocess
import pytest
from network_wrangler import ProjectCard
from network_wrangler import RoadwayNetwork
from network_wrangler import TransitNetwork
from network_wrangler import Scenario
from network_wrangler.logger import WranglerLogger

"""
Run just the tests labeled scenario using `pytest -v -m scenario`
To run with print statments, use `pytest -s -m scenario`
"""

STPAUL_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.realpath(__file__))), "examples", "stpaul"
)
SCRATCH_DIR = os.path.join(os.getcwd(), "scratch")

STPAUL_SHAPE_FILE = os.path.join(STPAUL_DIR, "shape.geojson")
STPAUL_LINK_FILE = os.path.join(STPAUL_DIR, "link.json")
STPAUL_NODE_FILE = os.path.join(STPAUL_DIR, "node.geojson")


@pytest.mark.scenario
@pytest.mark.travis
def test_project_card_read(request):
    print("\n--Starting:", request.node.name)
    in_dir = os.path.join(
        os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
        "examples",
        "stpaul",
        "project_cards",
    )
    in_file = os.path.join(in_dir, "1_simple_roadway_attribute_change.yml")
    project_card = ProjectCard.read(in_file)
    WranglerLogger.info(project_card)
    print(str(project_card))
    assert project_card.category == "Roadway Property Change"
    print("--Finished:", request.node.name)


@pytest.mark.scenario
@pytest.mark.travis
def test_project_card_write(request):
    print("\n--Starting:", request.node.name)
    in_dir = os.path.join(STPAUL_DIR, "project_cards")
    in_file = os.path.join(in_dir, "1_simple_roadway_attribute_change.yml")
    outfile = os.path.join(SCRATCH_DIR, "t_simple_roadway_attribute_change.yml")
    project_card = ProjectCard.read(in_file)
    project_card.write(outfile)
    test_card = ProjectCard.read(in_file)
    for k, v in project_card.__dict__.items():
        assert v == test_card.__dict__[k]


@pytest.mark.scenario
@pytest.mark.travis
def test_scenario_conflicts(request):

    project_cards_list = []
    project_cards_list.append(
        ProjectCard.read(
            os.path.join(STPAUL_DIR, "project_cards", "a_test_project_card.yml")
        )
    )
    project_cards_list.append(
        ProjectCard.read(
            os.path.join(STPAUL_DIR, "project_cards", "b_test_project_card.yml")
        )
    )
    project_cards_list.append(
        ProjectCard.read(
            os.path.join(STPAUL_DIR, "project_cards", "c_test_project_card.yml")
        )
    )

    scen = Scenario.create_scenario(
        base_scenario={}, project_cards_list=project_cards_list
    )

    print(str(scen), "\n")

    scen.check_scenario_conflicts()
    if scen.has_conflict_error:
        print("Conflicting project found for scenario!")

    print("Conflict checks done:", scen.conflicts_checked)
    print("--Finished:", request.node.name)


@pytest.mark.scenario
@pytest.mark.travis
def test_scenario_requisites(request):
    print("\n--Starting:", request.node.name)
    base_scenario = {}

    project_cards_list = []
    project_cards_list.append(
        ProjectCard.read(
            os.path.join(STPAUL_DIR, "project_cards", "a_test_project_card.yml")
        )
    )
    project_cards_list.append(
        ProjectCard.read(
            os.path.join(STPAUL_DIR, "project_cards", "b_test_project_card.yml")
        )
    )
    project_cards_list.append(
        ProjectCard.read(
            os.path.join(STPAUL_DIR, "project_cards", "c_test_project_card.yml")
        )
    )

    scen = Scenario.create_scenario(
        base_scenario=base_scenario, project_cards_list=project_cards_list
    )

    print(str(scen), "\n")

    scen.check_scenario_requisites()
    if scen.has_requisite_error:
        print("Missing pre- or co-requisite projects found for scenario!")

    print("Requisite checks done:", scen.requisites_checked)
    print("--Finished:", request.node.name)


@pytest.mark.scenario
@pytest.mark.travis
def test_project_sort(request):
    print("\n--Starting:", request.node.name)
    base_scenario = {}

    project_cards_list = []
    project_cards_list.append(
        ProjectCard.read(
            os.path.join(STPAUL_DIR, "project_cards", "a_test_project_card.yml")
        )
    )
    project_cards_list.append(
        ProjectCard.read(
            os.path.join(STPAUL_DIR, "project_cards", "b_test_project_card.yml")
        )
    )
    project_cards_list.append(
        ProjectCard.read(
            os.path.join(STPAUL_DIR, "project_cards", "c_test_project_card.yml")
        )
    )

    scen = Scenario.create_scenario(
        base_scenario=base_scenario, project_cards_list=project_cards_list
    )
    print("\n> Prerequisites:")
    import pprint

    pprint.pprint(scen.prerequisites)
    print("\nUnordered Projects:", scen.get_project_names())
    scen.check_scenario_conflicts()
    scen.check_scenario_requisites()

    scen.order_project_cards()
    print("Ordered Projects:", scen.get_project_names())
    print("--Finished:", request.node.name)


@pytest.mark.roadway
@pytest.mark.scenario
@pytest.mark.travis
def test_managed_lane_project_card(request):
    print("\n--Starting:", request.node.name)

    print("Reading project card ...")
    project_card_name = "5_managed_lane.yml"
    project_card_path = os.path.join(
        os.getcwd(), "examples", "stpaul", "project_cards", project_card_name
    )
    project_card = ProjectCard.read(project_card_path)
    print(project_card)

    print("--Finished:", request.node.name)


# selection, answer
query_tests = [
    # TEST 1
    (
        # SELECTION 1
        {
            "selection": {
                "link": [{"name": ["6th", "Sixth", "sixth"]}],
                "A": {"osm_node_id": "187899923"},  # start searching for segments at A
                "B": {"osm_node_id": "187865924"},  # end at B
            },
            "ignore": [],
        },
        # ANSWER 1
        '((name.str.contains("6th") or '
        + 'name.str.contains("Sixth") or '
        + 'name.str.contains("sixth")) and '
        + "(drive_access==1))",
    ),
    # TEST 2
    (
        # SELECTION 2
        {
            "selection": {
                "link": [{"name": ["6th", "Sixth", "sixth"]}],
                "A": {"osm_node_id": "187899923"},  # start searching for segments at A
                "B": {"osm_node_id": "187865924"},  # end at B
            },
            "ignore": ["name"],
        },
        # ANSWER 1
        "((drive_access==1))",
    ),
    # TEST 3
    (
        # SELECTION 3
        {
            "selection": {
                "link": [
                    {
                        "name": ["6th", "Sixth", "sixth"]
                    },  # find streets that have one of the various forms of 6th
                    {"lanes": [1, 2]},  # only select links that are either 1 or 2 lanes
                    {
                        "bike_access": [1]
                    },  # only select links that are marked for biking
                ],
                "A": {"osm_node_id": "187899923"},  # start searching for segments at A
                "B": {"osm_node_id": "187865924"},  # end at B
            },
            "ignore": [],
        },
        # ANSWER 3
        '((name.str.contains("6th") or '
        + 'name.str.contains("Sixth") or '
        + 'name.str.contains("sixth")) and '
        + "(lanes==1 or lanes==2) and "
        + "(bike_access==1) and (drive_access==1))",
    ),
    # TEST 4
    (
        # SELECTION 4
        {
            "selection": {
                "link": [
                    {
                        "name": ["6th", "Sixth", "sixth"]
                    },  # find streets that have one of the various forms of 6th
                    {"model_link_id": [134574]},
                    {"lanes": [1, 2]},  # only select links that are either 1 or 2 lanes
                    {
                        "bike_access": [1]
                    },  # only select links that are marked for biking
                ],
                "A": {"osm_node_id": "187899923"},  # start searching for segments at A
                "B": {"osm_node_id": "187865924"},  # end at B
            },
            "ignore": [],
        },
        # ANSWER 4
        "((model_link_id==134574))",
    ),
]


@pytest.mark.parametrize("test_spec", query_tests)
@pytest.mark.travis
def test_query_builder(request, test_spec):
    selection, answer = test_spec

    sel_query = ProjectCard.build_link_selection_query(
        selection=selection["selection"],
        unique_model_link_identifiers=RoadwayNetwork.UNIQUE_MODEL_LINK_IDENTIFIERS,
        ignore=selection["ignore"],
    )

    print("\nsel_query:\n", sel_query)
    print("\nanswer:\n", answer)
    assert sel_query == answer

    print("--Finished:", request.node.name)


@pytest.mark.scenario
@pytest.mark.travis
def test_apply_summary_wrappers(request):
    print("\n--Starting:", request.node.name)

    card_filenames = [
        "3_multiple_roadway_attribute_change.yml",
        "multiple_changes.yml",
        "4_simple_managed_lane.yml",
    ]

    project_card_directory = os.path.join(STPAUL_DIR, "project_cards")

    project_cards_list = [
        ProjectCard.read(os.path.join(project_card_directory, filename), validate=False)
        for filename in card_filenames
    ]

    base_scenario = {
        "road_net": RoadwayNetwork.read(
            link_file=STPAUL_LINK_FILE,
            node_file=STPAUL_NODE_FILE,
            shape_file=STPAUL_SHAPE_FILE,
            fast=True,
        ),
        "transit_net": TransitNetwork.read(STPAUL_DIR),
    }

    my_scenario = Scenario.create_scenario(
        base_scenario=base_scenario, project_cards_list=project_cards_list
    )

    my_scenario.apply_all_projects()

    my_scenario.scenario_summary()

    print("--Finished:", request.node.name)


@pytest.mark.scenario
@pytest.mark.travis
def test_scenario_building_from_script(request):
    print("\n--Starting:", request.node.name)

    config_file = os.path.join(os.getcwd(), "examples", "config_1.yml")
    # config_file = os.path.join(os.getcwd(),"example","config_2.yml")
    script_to_run = os.path.join(os.getcwd(), "scripts", "build_scenario.py")

    # replace backward slash with forward slash
    config_file = config_file.replace(os.sep, "/")
    script_to_run = script_to_run.replace(os.sep, "/")

    # print(config_file)
    # print(script_to_run)

    p = subprocess.Popen([sys.executable, script_to_run, config_file])
    p.communicate()  # wait for the subprocess call to finish

    print("--Finished:", request.node.name)



from setuptools import setup

classifiers = [
    "Development Status :: 1 - Planning",
    "License :: OSI Approved :: Apache Software License",
    "Natural Language :: English",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
]

with open("README.md") as f:
    long_description = f.read()

with open("requirements.txt") as f:
    requirements = f.readlines()

with open("dev-requirements.txt") as f:
    requirements += f.readlines()

install_requires = [r.strip() for r in requirements]

setup(
    name="network_wrangler",
    version="0.2.0",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wsp-sag/network_wrangler",
    license="Apache 2",
    platforms="any",
    packages=["network_wrangler"],
    include_package_data=True,
    install_requires=install_requires,
)
